@@ -41,9 +41,9 @@ type DNSContext struct {
 	// localIP - local IP address (for UDP socket to call udpMakeOOBWithSrc)
 	localIP net.IP
 
-	// HTTPRequest - HTTP request (for DOH only)
+	// HTTPRequest - HTTP request (for DoH only)
 	HTTPRequest *http.Request
-	// HTTPResponseWriter - HTTP response writer (for DOH only)
+	// HTTPResponseWriter - HTTP response writer (for DoH only)
 	HTTPResponseWriter http.ResponseWriter
 
 	// DNSCryptResponseWriter - necessary to respond to a DNSCrypt query
@@ -53,9 +53,14 @@ type DNSContext struct {
 	// ProtoQUIC only.
 	QUICStream quic.Stream
 
-	// QUICSession is the QUIC session from which we got the query.  For
+	// QUICConnection is the QUIC session from which we got the query.  For
 	// ProtoQUIC only.
-	QUICSession quic.Session
+	QUICConnection quic.Connection
+
+	// DoQVersion is the DoQ protocol version. It can (and should) be read from
+	// ALPN, but in the current version we also use the way DNS messages are
+	// encoded as a signal.
+	DoQVersion DoQVersion
 
 	// RequestID is an opaque numerical identifier of this request that is
 	// guaranteed to be unique across requests processed by a single Proxy
@@ -113,3 +118,17 @@ func (dctx *DNSContext) scrub() {
 	// Some devices require DNS message compression.
 	dctx.Res.Compress = true
 }
+
+// DoQVersion is an enumeration with supported DoQ versions.
+type DoQVersion int
+
+const (
+	// DoQv1Draft represents old DoQ draft versions that do not send a 2-octet
+	// prefix with the DNS message length.
+	//
+	// TODO(ameshkov): remove in the end of 2024.
+	DoQv1Draft DoQVersion = 0x00
+
+	// DoQv1 represents DoQ v1.0: https://www.rfc-editor.org/rfc/rfc9250.html.
+	DoQv1 DoQVersion = 0x01
+)
