@@ -9,6 +9,7 @@ import (
 	"net"
 	"net/http"
 	"net/url"
+	"runtime"
 	"sync"
 	"time"
 
@@ -47,8 +48,8 @@ type dnsOverHTTPS struct {
 	// The Client's Transport typically has internal state (cached TCP
 	// connections), so Clients should be reused instead of created as
 	// needed. Clients are safe for concurrent use by multiple goroutines.
-	client      *http.Client
-	clientGuard sync.Mutex
+	client   *http.Client
+	clientMu sync.Mutex
 
 	// quicConfig is the QUIC configuration that is used if HTTP/3 is enabled
 	// for this upstream.
@@ -57,7 +58,7 @@ type dnsOverHTTPS struct {
 }
 
 // type check
-var _ Upstream = &dnsOverHTTPS{}
+var _ Upstream = (*dnsOverHTTPS)(nil)
 
 // newDoH returns the DNS-over-HTTPS Upstream.
 func newDoH(uu *url.URL, opts *Options) (u Upstream, err error) {
@@ -69,14 +70,19 @@ func newDoH(uu *url.URL, opts *Options) (u Upstream, err error) {
 		return nil, fmt.Errorf("creating https bootstrapper: %w", err)
 	}
 
-	return &dnsOverHTTPS{
+	u = &dnsOverHTTPS{
 		boot: b,
 
 		quicConfig: &quic.Config{
 			KeepAlivePeriod: QUICKeepAlivePeriod,
 			TokenStore:      newQUICTokenStore(),
+			Tracer:          opts.QUICTracer,
 		},
-	}, nil
+	}
+
+	runtime.SetFinalizer(u, (*dnsOverHTTPS).Close)
+
+	return u, nil
 }
 
 // Address implements the Upstream interface for *dnsOverHTTPS.
@@ -101,42 +107,67 @@ func (p *dnsOverHTTPS) Exchange(m *dns.Msg) (resp *dns.Msg, err error) {
 
 	// Check if there was already an active client before sending the request.
 	// We'll only attempt to re-connect if there was one.
-	hasClient := p.hasClient()
+	client, isCached, err := p.getClient()
+	if err != nil {
+		return nil, fmt.Errorf("failed to init http client: %w", err)
+	}
 
 	// Make the first attempt to send the DNS query.
-	resp, err = p.exchangeHTTPS(m)
+	resp, err = p.exchangeHTTPS(client, m)
 
 	// Make up to 2 attempts to re-create the HTTP client and send the request
 	// again.  There are several cases (mostly, with QUIC) where this workaround
 	// is necessary to make HTTP client usable.  We need to make 2 attempts in
 	// the case when the connection was closed (due to inactivity for example)
 	// AND the server refuses to open a 0-RTT connection.
-	for i := 0; hasClient && p.shouldRetry(err) && i < 2; i++ {
-		log.Debug("re-creating the HTTP client and retrying due to %v", err)
+	for i := 0; isCached && p.shouldRetry(err) && i < 2; i++ {
+		client, err = p.resetClient(err)
+		if err != nil {
+			return nil, fmt.Errorf("failed to reset http client: %w", err)
+		}
 
-		// Make sure we reset the client here.
-		p.resetClient(err)
-
-		resp, err = p.exchangeHTTPS(m)
+		resp, err = p.exchangeHTTPS(client, m)
 	}
 
 	if err != nil {
 		// If the request failed anyway, make sure we don't use this client.
-		p.resetClient(err)
+		_, resErr := p.resetClient(err)
+
+		return nil, errors.WithDeferred(err, resErr)
 	}
 
 	return resp, err
 }
 
-// exchangeHTTPS creates an HTTP client and sends the DNS query using it.
-func (p *dnsOverHTTPS) exchangeHTTPS(m *dns.Msg) (resp *dns.Msg, err error) {
-	client, err := p.getClient()
-	if err != nil {
-		return nil, fmt.Errorf("initializing http client: %w", err)
+// Close implements the Upstream interface for *dnsOverHTTPS.
+func (p *dnsOverHTTPS) Close() (err error) {
+	p.clientMu.Lock()
+	defer p.clientMu.Unlock()
+
+	runtime.SetFinalizer(p, nil)
+
+	if p.client == nil {
+		return nil
 	}
 
-	logBegin(p.Address(), m)
-	resp, err = p.exchangeHTTPSClient(m, client)
+	return p.closeClient(p.client)
+}
+
+// closeClient cleans up resources used by client if necessary.  Note, that at
+// this point it should only be done for HTTP/3 as it may leak due to keep-alive
+// connections.
+func (p *dnsOverHTTPS) closeClient(client *http.Client) (err error) {
+	if isHTTP3(client) {
+		return client.Transport.(io.Closer).Close()
+	}
+
+	return nil
+}
+
+// exchangeHTTPS logs the request and its result and calls exchangeHTTPSClient.
+func (p *dnsOverHTTPS) exchangeHTTPS(client *http.Client, req *dns.Msg) (resp *dns.Msg, err error) {
+	logBegin(p.Address(), req)
+	resp, err = p.exchangeHTTPSClient(client, req)
 	logFinish(p.Address(), err)
 
 	return resp, err
@@ -144,8 +175,11 @@ func (p *dnsOverHTTPS) exchangeHTTPS(m *dns.Msg) (resp *dns.Msg, err error) {
 
 // exchangeHTTPSClient sends the DNS query to a DoH resolver using the specified
 // http.Client instance.
-func (p *dnsOverHTTPS) exchangeHTTPSClient(m *dns.Msg, client *http.Client) (*dns.Msg, error) {
-	buf, err := m.Pack()
+func (p *dnsOverHTTPS) exchangeHTTPSClient(
+	client *http.Client,
+	req *dns.Msg,
+) (resp *dns.Msg, err error) {
+	buf, err := req.Pack()
 	if err != nil {
 		return nil, fmt.Errorf("packing message: %w", err)
 	}
@@ -153,7 +187,7 @@ func (p *dnsOverHTTPS) exchangeHTTPSClient(m *dns.Msg, client *http.Client) (*dn
 	// It appears, that GET requests are more memory-efficient with Golang
 	// implementation of HTTP/2.
 	method := http.MethodGet
-	if _, ok := client.Transport.(*http3.RoundTripper); ok {
+	if isHTTP3(client) {
 		// If we're using HTTP/3, use http3.MethodGet0RTT to force using 0-RTT.
 		method = http3.MethodGet0RTT
 	}
@@ -165,50 +199,51 @@ func (p *dnsOverHTTPS) exchangeHTTPSClient(m *dns.Msg, client *http.Client) (*dn
 		RawQuery: fmt.Sprintf("dns=%s", base64.RawURLEncoding.EncodeToString(buf)),
 	}
 
-	req, err := http.NewRequest(method, u.String(), nil)
+	httpReq, err := http.NewRequest(method, u.String(), nil)
 	if err != nil {
 		return nil, fmt.Errorf("creating http request to %s: %w", p.boot.URL, err)
 	}
 
-	req.Header.Set("Accept", "application/dns-message")
-	req.Header.Set("User-Agent", "")
+	httpReq.Header.Set("Accept", "application/dns-message")
+	httpReq.Header.Set("User-Agent", "")
 
-	resp, err := client.Do(req)
-	if resp != nil && resp.Body != nil {
-		defer resp.Body.Close()
-	}
+	httpResp, err := client.Do(httpReq)
 	if err != nil {
 		return nil, fmt.Errorf("requesting %s: %w", p.boot.URL, err)
 	}
+	defer httpResp.Body.Close()
 
-	body, err := io.ReadAll(resp.Body)
+	body, err := io.ReadAll(httpResp.Body)
 	if err != nil {
 		return nil, fmt.Errorf("reading %s: %w", p.boot.URL, err)
 	}
 
-	if resp.StatusCode != http.StatusOK {
-		return nil, fmt.Errorf("got status code %d from %s", resp.StatusCode, p.boot.URL)
+	if httpResp.StatusCode != http.StatusOK {
+		return nil,
+			fmt.Errorf(
+				"expected status %d, got %d from %s",
+				http.StatusOK,
+				httpResp.StatusCode,
+				p.boot.URL,
+			)
 	}
 
-	response := dns.Msg{}
-	err = response.Unpack(body)
+	resp = &dns.Msg{}
+	err = resp.Unpack(body)
 	if err != nil {
-		return nil, fmt.Errorf("unpacking response from %s: body is %s: %w", p.boot.URL, string(body), err)
+		return nil, fmt.Errorf(
+			"unpacking response from %s: body is %s: %w",
+			p.boot.URL,
+			body,
+			err,
+		)
 	}
 
-	if response.Id != m.Id {
+	if resp.Id != req.Id {
 		err = dns.ErrId
 	}
 
-	return &response, err
-}
-
-// hasClient returns true if this connection already has an active HTTP client.
-func (p *dnsOverHTTPS) hasClient() (ok bool) {
-	p.clientGuard.Lock()
-	defer p.clientGuard.Unlock()
-
-	return p.client != nil
+	return resp, err
 }
 
 // shouldRetry checks what error we have received and returns true if we should
@@ -238,16 +273,27 @@ func (p *dnsOverHTTPS) shouldRetry(err error) (ok bool) {
 // resetClient triggers re-creation of the *http.Client that is used by this
 // upstream.  This method accepts the error that caused resetting client as
 // depending on the error we may also reset the QUIC config.
-func (p *dnsOverHTTPS) resetClient(err error) {
-	p.clientGuard.Lock()
-	defer p.clientGuard.Unlock()
+func (p *dnsOverHTTPS) resetClient(resetErr error) (client *http.Client, err error) {
+	p.clientMu.Lock()
+	defer p.clientMu.Unlock()
 
-	p.client = nil
-
-	if errors.Is(err, quic.Err0RTTRejected) {
+	if errors.Is(resetErr, quic.Err0RTTRejected) {
 		// Reset the TokenStore only if 0-RTT was rejected.
 		p.resetQUICConfig()
 	}
+
+	oldClient := p.client
+	if oldClient != nil {
+		closeErr := p.closeClient(oldClient)
+		if closeErr != nil {
+			log.Info("warning: failed to close the old http client: %v", closeErr)
+		}
+	}
+
+	log.Debug("re-creating the http client due to %v", resetErr)
+	p.client, err = p.createClient()
+
+	return p.client, err
 }
 
 // getQUICConfig returns the QUIC config in a thread-safe manner.  Note, that
@@ -271,25 +317,26 @@ func (p *dnsOverHTTPS) resetQUICConfig() {
 
 // getClient gets or lazily initializes an HTTP client (and transport) that will
 // be used for this DoH resolver.
-func (p *dnsOverHTTPS) getClient() (c *http.Client, err error) {
+func (p *dnsOverHTTPS) getClient() (c *http.Client, isCached bool, err error) {
 	startTime := time.Now()
 
-	p.clientGuard.Lock()
-	defer p.clientGuard.Unlock()
+	p.clientMu.Lock()
+	defer p.clientMu.Unlock()
 	if p.client != nil {
-		return p.client, nil
+		return p.client, true, nil
 	}
 
 	// Timeout can be exceeded while waiting for the lock. This happens quite
 	// often on mobile devices.
 	elapsed := time.Since(startTime)
 	if p.boot.options.Timeout > 0 && elapsed > p.boot.options.Timeout {
-		return nil, fmt.Errorf("timeout exceeded: %s", elapsed)
+		return nil, false, fmt.Errorf("timeout exceeded: %s", elapsed)
 	}
 
+	log.Debug("creating a new http client")
 	p.client, err = p.createClient()
 
-	return p.client, err
+	return p.client, false, err
 }
 
 // createClient creates a new *http.Client instance.  The HTTP protocol version
@@ -368,6 +415,53 @@ func (p *dnsOverHTTPS) createTransport() (t http.RoundTripper, err error) {
 	return transport, nil
 }
 
+// http3Transport is a wrapper over *http3.RoundTripper that tries to optimize
+// its behavior.  The main thing that it does is trying to force use a single
+// connection to a host instead of creating a new one all the time.  It also
+// helps mitigate race issues with quic-go.
+type http3Transport struct {
+	baseTransport *http3.RoundTripper
+
+	closed bool
+	mu     sync.RWMutex
+}
+
+// type check
+var _ http.RoundTripper = (*http3Transport)(nil)
+
+// RoundTrip implements the http.RoundTripper interface for *http3Transport.
+func (h *http3Transport) RoundTrip(req *http.Request) (resp *http.Response, err error) {
+	h.mu.RLock()
+	defer h.mu.RUnlock()
+
+	if h.closed {
+		return nil, net.ErrClosed
+	}
+
+	// Try to use cached connection to the target host if it's available.
+	resp, err = h.baseTransport.RoundTripOpt(req, http3.RoundTripOpt{OnlyCachedConn: true})
+
+	if errors.Is(err, http3.ErrNoCachedConn) {
+		// If there are no cached connection, trigger creating a new one.
+		resp, err = h.baseTransport.RoundTrip(req)
+	}
+
+	return resp, err
+}
+
+// type check
+var _ io.Closer = (*http3Transport)(nil)
+
+// Close implements the io.Closer interface for *http3Transport.
+func (h *http3Transport) Close() (err error) {
+	h.mu.Lock()
+	defer h.mu.Unlock()
+
+	h.closed = true
+
+	return h.baseTransport.Close()
+}
+
 // createTransportH3 tries to create an HTTP/3 transport for this upstream.
 // We should be able to fall back to H1/H2 in case if HTTP/3 is unavailable or
 // if it is too slow.  In order to do that, this method will run two probes
@@ -404,7 +498,7 @@ func (p *dnsOverHTTPS) createTransportH3(
 		QuicConfig:         p.getQUICConfig(),
 	}
 
-	return rt, nil
+	return &http3Transport{baseTransport: rt}, nil
 }
 
 // probeH3 runs a test to check whether QUIC is faster than TLS for this
@@ -553,3 +647,10 @@ func (p *dnsOverHTTPS) supportedHTTPVersions() (v []HTTPVersion) {
 
 	return v
 }
+
+// isHTTP3 checks if the *http.Client is an HTTP/3 client.
+func isHTTP3(client *http.Client) (ok bool) {
+	_, ok = client.Transport.(*http3Transport)
+
+	return ok
+}
