@@ -5,6 +5,7 @@ import (
 	"fmt"
 	"net"
 	"net/url"
+	"runtime"
 	"sync"
 	"time"
 
@@ -49,8 +50,8 @@ type dnsOverQUIC struct {
 
 	// conn is the current active QUIC connection.  It can be closed and
 	// re-opened when needed.
-	conn      quic.Connection
-	connGuard sync.RWMutex
+	conn   quic.Connection
+	connMu sync.RWMutex
 
 	// bytesPool is a *sync.Pool we use to store byte buffers in.  These byte
 	// buffers are used to read responses from the upstream.
@@ -59,7 +60,7 @@ type dnsOverQUIC struct {
 }
 
 // type check
-var _ Upstream = &dnsOverQUIC{}
+var _ Upstream = (*dnsOverQUIC)(nil)
 
 // newDoQ returns the DNS-over-QUIC Upstream.
 func newDoQ(uu *url.URL, opts *Options) (u Upstream, err error) {
@@ -71,13 +72,18 @@ func newDoQ(uu *url.URL, opts *Options) (u Upstream, err error) {
 		return nil, fmt.Errorf("creating quic bootstrapper: %w", err)
 	}
 
-	return &dnsOverQUIC{
+	u = &dnsOverQUIC{
 		boot: b,
 		quicConfig: &quic.Config{
 			KeepAlivePeriod: QUICKeepAlivePeriod,
 			TokenStore:      newQUICTokenStore(),
+			Tracer:          opts.QUICTracer,
 		},
-	}, nil
+	}
+
+	runtime.SetFinalizer(u, (*dnsOverQUIC).Close)
+
+	return u, nil
 }
 
 // Address implements the Upstream interface for *dnsOverQUIC.
@@ -128,6 +134,20 @@ func (p *dnsOverQUIC) Exchange(m *dns.Msg) (resp *dns.Msg, err error) {
 	return resp, err
 }
 
+// Close implements the Upstream interface for *dnsOverQUIC.
+func (p *dnsOverQUIC) Close() (err error) {
+	p.connMu.Lock()
+	defer p.connMu.Unlock()
+
+	runtime.SetFinalizer(p, nil)
+
+	if p.conn != nil {
+		err = p.conn.CloseWithError(QUICCodeNoError, "")
+	}
+
+	return err
+}
+
 // exchangeQUIC attempts to open a QUIC connection, send the DNS message
 // through it and return the response it got from the server.
 func (p *dnsOverQUIC) exchangeQUIC(m *dns.Msg) (resp *dns.Msg, err error) {
@@ -193,10 +213,10 @@ func (p *dnsOverQUIC) getBytesPool() (pool *sync.Pool) {
 // close the existing one if needed.
 func (p *dnsOverQUIC) getConnection(useCached bool) (quic.Connection, error) {
 	var conn quic.Connection
-	p.connGuard.RLock()
+	p.connMu.RLock()
 	conn = p.conn
 	if conn != nil && useCached {
-		p.connGuard.RUnlock()
+		p.connMu.RUnlock()
 
 		return conn, nil
 	}
@@ -204,10 +224,10 @@ func (p *dnsOverQUIC) getConnection(useCached bool) (quic.Connection, error) {
 		// we're recreating the connection, let's create a new one.
 		_ = conn.CloseWithError(QUICCodeNoError, "")
 	}
-	p.connGuard.RUnlock()
+	p.connMu.RUnlock()
 
-	p.connGuard.Lock()
-	defer p.connGuard.Unlock()
+	p.connMu.Lock()
+	defer p.connMu.Unlock()
 
 	var err error
 	conn, err = p.openConnection()
@@ -221,8 +241,8 @@ func (p *dnsOverQUIC) getConnection(useCached bool) (quic.Connection, error) {
 
 // hasConnection returns true if there's an active QUIC connection.
 func (p *dnsOverQUIC) hasConnection() (ok bool) {
-	p.connGuard.Lock()
-	defer p.connGuard.Unlock()
+	p.connMu.Lock()
+	defer p.connMu.Unlock()
 
 	return p.conn != nil
 }
@@ -305,8 +325,8 @@ func (p *dnsOverQUIC) openConnection() (conn quic.Connection, err error) {
 // new queries were processed in another connection.  We can do that in the case
 // of a fatal error.
 func (p *dnsOverQUIC) closeConnWithError(err error) {
-	p.connGuard.Lock()
-	defer p.connGuard.Unlock()
+	p.connMu.Lock()
+	defer p.connMu.Unlock()
 
 	if p.conn == nil {
 		// Do nothing, there's no active conn anyways.
@@ -393,6 +413,25 @@ func isQUICRetryError(err error) (ok bool) {
 		return true
 	}
 
+	var resetErr *quic.StatelessResetError
+	if errors.As(err, &resetErr) {
+		// A stateless reset is sent when a server receives a QUIC packet that
+		// it doesn't know how to decrypt.  For instance, it may happen when
+		// the server was recently rebooted.  We should reconnect and try again
+		// in this case.
+		return true
+	}
+
+	var qTransportError *quic.TransportError
+	if errors.As(err, &qTransportError) && qTransportError.ErrorCode == quic.NoError {
+		// A transport error with the NO_ERROR error code could be sent by the
+		// server when it considers that it's time to close the connection.
+		// For example, Google DNS eventually closes an active connection with
+		// the NO_ERROR code and "Connection max age expired" message:
+		// https://github.com/AdguardTeam/dnsproxy/issues/283
+		return true
+	}
+
 	if errors.Is(err, quic.Err0RTTRejected) {
 		// This error happens when we try to establish a 0-RTT connection with
 		// a token the server is no more aware of.  This can be reproduced by
